@@ -16,11 +16,16 @@
  *
  */
 
+#include <grpc/support/port_platform.h>
+
 #include "src/core/ext/filters/client_channel/lb_policy/grpclb/client_load_reporting_filter.h"
 
+#include <string.h>
+
 #include <grpc/support/atm.h>
 #include <grpc/support/log.h>
 
+#include "src/core/ext/filters/client_channel/lb_policy/grpclb/grpclb.h"
 #include "src/core/ext/filters/client_channel/lb_policy/grpclb/grpclb_client_stats.h"
 #include "src/core/lib/iomgr/error.h"
 #include "src/core/lib/profiling/timers.h"
@@ -33,18 +38,20 @@ static grpc_error* init_channel_elem(grpc_channel_element* elem,
 static void destroy_channel_elem(grpc_channel_element* elem) {}
 
 namespace {
+
 struct call_data {
   // Stats object to update.
-  grpc_grpclb_client_stats* client_stats;
+  grpc_core::RefCountedPtr<grpc_core::GrpcLbClientStats> client_stats;
   // State for intercepting send_initial_metadata.
   grpc_closure on_complete_for_send;
   grpc_closure* original_on_complete_for_send;
-  bool send_initial_metadata_succeeded;
+  bool send_initial_metadata_succeeded = false;
   // State for intercepting recv_initial_metadata.
   grpc_closure recv_initial_metadata_ready;
   grpc_closure* original_recv_initial_metadata_ready;
-  bool recv_initial_metadata_succeeded;
+  bool recv_initial_metadata_succeeded = false;
 };
+
 }  // namespace
 
 static void on_complete_for_send(void* arg, grpc_error* error) {
@@ -66,16 +73,8 @@ static void recv_initial_metadata_ready(void* arg, grpc_error* error) {
 
 static grpc_error* init_call_elem(grpc_call_element* elem,
                                   const grpc_call_element_args* args) {
-  call_data* calld = static_cast<call_data*>(elem->call_data);
-  // Get stats object from context and take a ref.
   GPR_ASSERT(args->context != nullptr);
-  if (args->context[GRPC_GRPCLB_CLIENT_STATS].value != nullptr) {
-    calld->client_stats =
-        grpc_grpclb_client_stats_ref(static_cast<grpc_grpclb_client_stats*>(
-            args->context[GRPC_GRPCLB_CLIENT_STATS].value));
-    // Record call started.
-    grpc_grpclb_client_stats_add_call_started(calld->client_stats);
-  }
+  new (elem->call_data) call_data();
   return GRPC_ERROR_NONE;
 }
 
@@ -86,38 +85,58 @@ static void destroy_call_elem(grpc_call_element* elem,
   if (calld->client_stats != nullptr) {
     // Record call finished, optionally setting client_failed_to_send and
     // received.
-    grpc_grpclb_client_stats_add_call_finished(
+    calld->client_stats->AddCallFinished(
         !calld->send_initial_metadata_succeeded /* client_failed_to_send */,
-        calld->recv_initial_metadata_succeeded /* known_received */,
-        calld->client_stats);
-    // All done, so unref the stats object.
-    grpc_grpclb_client_stats_unref(calld->client_stats);
+        calld->recv_initial_metadata_succeeded /* known_received */);
   }
+  calld->~call_data();
 }
 
 static void start_transport_stream_op_batch(
     grpc_call_element* elem, grpc_transport_stream_op_batch* batch) {
   call_data* calld = static_cast<call_data*>(elem->call_data);
   GPR_TIMER_SCOPE("clr_start_transport_stream_op_batch", 0);
-  if (calld->client_stats != nullptr) {
-    // Intercept send_initial_metadata.
-    if (batch->send_initial_metadata) {
-      calld->original_on_complete_for_send = batch->on_complete;
-      GRPC_CLOSURE_INIT(&calld->on_complete_for_send, on_complete_for_send,
-                        calld, grpc_schedule_on_exec_ctx);
-      batch->on_complete = &calld->on_complete_for_send;
+  // Handle send_initial_metadata.
+  if (batch->send_initial_metadata) {
+    // Grab client stats object from metadata.
+    grpc_linked_mdelem* client_stats_md =
+        batch->payload->send_initial_metadata.send_initial_metadata->list.head;
+    for (; client_stats_md != nullptr;
+         client_stats_md = client_stats_md->next) {
+      if (GRPC_SLICE_START_PTR(GRPC_MDKEY(client_stats_md->md)) ==
+          static_cast<const void*>(grpc_core::kGrpcLbClientStatsMetadataKey)) {
+        break;
+      }
     }
-    // Intercept recv_initial_metadata.
-    if (batch->recv_initial_metadata) {
-      calld->original_recv_initial_metadata_ready =
-          batch->payload->recv_initial_metadata.recv_initial_metadata_ready;
-      GRPC_CLOSURE_INIT(&calld->recv_initial_metadata_ready,
-                        recv_initial_metadata_ready, calld,
-                        grpc_schedule_on_exec_ctx);
-      batch->payload->recv_initial_metadata.recv_initial_metadata_ready =
-          &calld->recv_initial_metadata_ready;
+    if (client_stats_md != nullptr) {
+      grpc_core::GrpcLbClientStats* client_stats =
+          const_cast<grpc_core::GrpcLbClientStats*>(
+              reinterpret_cast<const grpc_core::GrpcLbClientStats*>(
+                  GRPC_SLICE_START_PTR(GRPC_MDVALUE(client_stats_md->md))));
+      if (client_stats != nullptr) {
+        calld->client_stats.reset(client_stats);
+        // Intercept completion.
+        calld->original_on_complete_for_send = batch->on_complete;
+        GRPC_CLOSURE_INIT(&calld->on_complete_for_send, on_complete_for_send,
+                          calld, grpc_schedule_on_exec_ctx);
+        batch->on_complete = &calld->on_complete_for_send;
+      }
+      // Remove metadata so it doesn't go out on the wire.
+      grpc_metadata_batch_remove(
+          batch->payload->send_initial_metadata.send_initial_metadata,
+          client_stats_md);
     }
   }
+  // Intercept completion of recv_initial_metadata.
+  if (batch->recv_initial_metadata) {
+    calld->original_recv_initial_metadata_ready =
+        batch->payload->recv_initial_metadata.recv_initial_metadata_ready;
+    GRPC_CLOSURE_INIT(&calld->recv_initial_metadata_ready,
+                      recv_initial_metadata_ready, calld,
+                      grpc_schedule_on_exec_ctx);
+    batch->payload->recv_initial_metadata.recv_initial_metadata_ready =
+        &calld->recv_initial_metadata_ready;
+  }
   // Chain to next filter.
   grpc_call_next_op(elem, batch);
 }
