@@ -19,10 +19,15 @@
 #ifndef GRPC_CORE_EXT_FILTERS_CLIENT_CHANNEL_RESOLVER_H
 #define GRPC_CORE_EXT_FILTERS_CLIENT_CHANNEL_RESOLVER_H
 
+#include <grpc/support/port_platform.h>
+
 #include <grpc/impl/codegen/grpc_types.h>
 
+#include "src/core/ext/filters/client_channel/server_address.h"
+#include "src/core/ext/filters/client_channel/service_config.h"
 #include "src/core/lib/gprpp/abstract.h"
 #include "src/core/lib/gprpp/orphanable.h"
+#include "src/core/lib/gprpp/ref_counted_ptr.h"
 #include "src/core/lib/iomgr/combiner.h"
 #include "src/core/lib/iomgr/iomgr.h"
 
@@ -42,25 +47,52 @@ namespace grpc_core {
 ///
 /// Note: All methods with a "Locked" suffix must be called from the
 /// combiner passed to the constructor.
-class Resolver : public InternallyRefCountedWithTracing<Resolver> {
+class Resolver : public InternallyRefCounted<Resolver> {
  public:
+  /// Results returned by the resolver.
+  struct Result {
+    ServerAddressList addresses;
+    RefCountedPtr<ServiceConfig> service_config;
+    grpc_error* service_config_error = GRPC_ERROR_NONE;
+    const grpc_channel_args* args = nullptr;
+
+    // TODO(roth): Remove everything below once grpc_error and
+    // grpc_channel_args are convert to copyable and movable C++ objects.
+    Result() = default;
+    ~Result();
+    Result(const Result& other);
+    Result(Result&& other);
+    Result& operator=(const Result& other);
+    Result& operator=(Result&& other);
+  };
+
+  /// A proxy object used by the resolver to return results to the
+  /// client channel.
+  class ResultHandler {
+   public:
+    virtual ~ResultHandler() {}
+
+    /// Returns a result to the channel.
+    /// Takes ownership of \a result.args.
+    virtual void ReturnResult(Result result) GRPC_ABSTRACT;  // NOLINT
+
+    /// Returns a transient error to the channel.
+    /// If the resolver does not set the GRPC_ERROR_INT_GRPC_STATUS
+    /// attribute on the error, calls will be failed with status UNKNOWN.
+    virtual void ReturnError(grpc_error* error) GRPC_ABSTRACT;
+
+    // TODO(yashkt): As part of the service config error handling
+    // changes, add a method to parse the service config JSON string.
+
+    GRPC_ABSTRACT_BASE_CLASS
+  };
+
   // Not copyable nor movable.
   Resolver(const Resolver&) = delete;
   Resolver& operator=(const Resolver&) = delete;
 
-  /// Requests a callback when a new result becomes available.
-  /// When the new result is available, sets \a *result to the new result
-  /// and schedules \a on_complete for execution.
-  /// If resolution is fatally broken, sets \a *result to nullptr and
-  /// schedules \a on_complete with an error.
-  ///
-  /// Note that the client channel will almost always have a request
-  /// to \a NextLocked() pending.  When it gets the callback, it will
-  /// process the new result and then immediately make another call to
-  /// \a NextLocked().  This allows push-based resolvers to provide new
-  /// data as soon as it becomes available.
-  virtual void NextLocked(grpc_channel_args** result,
-                          grpc_closure* on_complete) GRPC_ABSTRACT;
+  /// Starts resolving.
+  virtual void StartLocked() GRPC_ABSTRACT;
 
   /// Asks the resolver to obtain an updated resolver result, if
   /// applicable.
@@ -73,53 +105,47 @@ class Resolver : public InternallyRefCountedWithTracing<Resolver> {
   ///
   /// For push-based implementations, this may be a no-op.
   ///
-  /// If this causes new data to become available, then the currently
-  /// pending call to \a NextLocked() will return the new result.
-  ///
-  /// Note: Currently, all resolvers are required to return a new result
-  /// shortly after this method is called.  For pull-based mechanisms, if
-  /// the implementation decides to delay querying the name service, it
-  /// should immediately return a new copy of the previously returned
-  /// result (and it can then return the updated data later, when it
-  /// actually does query the name service).  For push-based mechanisms,
-  /// the implementation should immediately return a new copy of the
-  /// last-seen result.
-  /// TODO(roth): Remove this requirement once we fix pick_first to not
-  /// throw away unselected subchannels.
-  virtual void RequestReresolutionLocked() GRPC_ABSTRACT;
+  /// Note: Implementations must not invoke any method on the
+  /// ResultHandler from within this call.
+  virtual void RequestReresolutionLocked() {}
 
+  /// Resets the re-resolution backoff, if any.
+  /// This needs to be implemented only by pull-based implementations;
+  /// for push-based implementations, it will be a no-op.
+  /// TODO(roth): Pull the backoff code out of resolver and into
+  /// client_channel, so that it can be shared across resolver
+  /// implementations.  At that point, this method can go away.
+  virtual void ResetBackoffLocked() {}
+
+  // Note: This must be invoked while holding the combiner.
   void Orphan() override {
-    // Invoke ShutdownAndUnrefLocked() inside of the combiner.
-    GRPC_CLOSURE_SCHED(
-        GRPC_CLOSURE_CREATE(&Resolver::ShutdownAndUnrefLocked, this,
-                            grpc_combiner_scheduler(combiner_)),
-        GRPC_ERROR_NONE);
+    ShutdownLocked();
+    Unref();
   }
 
   GRPC_ABSTRACT_BASE_CLASS
 
  protected:
+  GRPC_ALLOW_CLASS_TO_USE_NON_PUBLIC_DELETE
+
   /// Does NOT take ownership of the reference to \a combiner.
   // TODO(roth): Once we have a C++-like interface for combiners, this
   // API should change to take a RefCountedPtr<>, so that we always take
   // ownership of a new ref.
-  explicit Resolver(grpc_combiner* combiner);
+  explicit Resolver(grpc_combiner* combiner,
+                    UniquePtr<ResultHandler> result_handler);
 
   virtual ~Resolver();
 
-  /// Shuts down the resolver.  If there is a pending call to
-  /// NextLocked(), the callback will be scheduled with an error.
+  /// Shuts down the resolver.
   virtual void ShutdownLocked() GRPC_ABSTRACT;
 
   grpc_combiner* combiner() const { return combiner_; }
 
- private:
-  static void ShutdownAndUnrefLocked(void* arg, grpc_error* ignored) {
-    Resolver* resolver = static_cast<Resolver*>(arg);
-    resolver->ShutdownLocked();
-    resolver->Unref();
-  }
+  ResultHandler* result_handler() const { return result_handler_.get(); }
 
+ private:
+  UniquePtr<ResultHandler> result_handler_;
   grpc_combiner* combiner_;
 };
 
