@@ -19,9 +19,11 @@
 #ifndef GRPC_CORE_LIB_SLICE_B64_H
 #define GRPC_CORE_LIB_SLICE_B64_H
 
+#include <grpc/support/port_platform.h>
+
 #include <grpc/slice.h>
 
-/* Encodes data using base64. It is the caller's responsability to free
+/* Encodes data using base64. It is the caller's responsibility to free
    the returned char * using gpr_free. Returns NULL on NULL input.
    TODO(makdharma) : change the flags to bool from int */
 char* grpc_base64_encode(const void* data, size_t data_size, int url_safe,
@@ -33,7 +35,7 @@ size_t grpc_base64_estimate_encoded_size(size_t data_size, int url_safe,
                                          int multiline);
 
 /* Encodes data using base64 and write it to memory pointed to by result. It is
- * the caller's responsiblity to allocate enough memory in |result| to fit the
+ * the caller's responsibility to allocate enough memory in |result| to fit the
  * encoded data. */
 void grpc_base64_encode_core(char* result, const void* vdata, size_t data_size,
                              int url_safe, int multiline);
